/****************************************************************************
 * 
 *      Copyright (c) 1991  Corporation of the President of 
 *                          the CHURCH of JESUS CHRIST of LATTER-DAY SAINTS.
 *
 *      All rights reserved.  Unauthorized reproduction of this software is
 *      prohibited and is in violaton of United States copyright laws.
 * 
 ****************************************************************************/
/*******************************************************************
 *
 * FILE: GEDDISK.C  
 * 
 * DESCRIPTION:  Part of the hash package which is included in the GEDCOM
 * library.
 * 
 * FUNCTIONS IN THIS FILE:
 * 
 *  GHBool gedh_closer_fit(void)
 *  int gedh_delete_disk(HASH_TAB *htab, char * key, KSIZ keyLen)
 *  void *  gedh_dlookup
 *      (HASH_TAB *htab, void *key, KSIZ keyLen, void *val, LEN *valLen, int *status)
 *  void ged_elseek( int handle, long offset, int origin)
 *  int ged_eread( int handle, void *buffer, unsigned count )
 *  void gedh_getRecordType(void)
 *  void gedh_hash_seek(void)
 *  int gedh_insertDisk
 *      (HASH_TAB *htab, char *key, KSIZ keyLen, char *val, LEN valLen)
 *  int gedh_iterate_disk
 *      (HASH_TAB *htab, char *key, KSIZ *keyLen, char *val, LEN *valLen)
 *  LEN gedh_kvsize(KVDISK *kv)
 *  void gedh_loadKv(void)
 *  void gedh_loadNxtKv(void)
 *  void gedh_makeGarb(void)
 *  void gedh_read_block(char *buf)
 *  int gedh_save_htab_struct(HASH_TAB *htab)
 *  void gedh_setGarb(void)
 *  void gedh_toBuf(void)
 *  void gedh_toGarbBuf(void)
 *  void gedh_toNxtBlk(void)
 *  void gedh_toOldBuf(void)
 *  void gedh_write_block(void)
 *  void gedh_writeEOL(void)
 *  void gedh_writeNewBlk(void)
 * 
 *
 * MODIFICATION HISTORY:
 * 
 * ----------------------------------------------------------------------
 * |  Date           | Who | Comments 
 * ----------------------------------------------------------------------
 * | 08/23/91 08:11am| LAM | No modifications needed.  Code was perfect from day one |-)
 * ----------------------------------------------------------------------
 ********************************************************************/

#include "gedcom.h"
#include "GEDHSTAT.H" /* Static declarations used internally by library functions */
#include "gedisk.h"

/****************************************************************************
 *
 * NAME: gedh_delete_disk
 *
 * DESCRIPTION: Finds the key in disk-based hash table and deletes its record by
 * turning it into garbage.
 * 
 * RETURN:  Zero if key not found, else SUCCESS.
 *
 * CALLED BY: gedhash_delete
 *
 * PARAMETERS:
 *
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: None
 *
 * key:   Input: If this key is found, record will be deleted from hash table
 *       output: None
 *
 * keyLen: Input: Length in bytes of key.
 *
 ****************************************************************************/
int gedh_delete_disk(HASH_TAB *htab, char * key, KSIZ keyLen)
  {
  State state =  LSEEK_TO_HASH;  /* First state to start off with */

  /* Initialize global variables */
  Htab = htab;
  Key = key;
  Htab = htab;
  Key = key;
  KeyLen = keyLen;
  Fh = htab->fh;
  Garbage.block = 0L; /* Initialize Garbage struct to 0 */
  Garbage.garb_record = NULL;
  Garbage.size = (LEN)0;
  Filepos = 0L;
  Buffer_change_flag = FALSE;
  Return_status = SUCCESS;

  WHILESTATE(state)
    {
    STATE LSEEK_TO_HASH :  /* lseek to right block in hash table & read in value */
      COND(TRUE, gedh_hash_seek, LOAD_KV);

    STATE GET_NEXT_KV   :  /* Jump to next kv record and read from buffer */
      COND(TRUE, gedh_loadNxtKv, GET_RECORD_TYPE);

    STATE LOAD_KV       :  /* Like GET_NEXT_KV, except don't jump to next kv record. */
      COND( TRUE, gedh_loadKv, GET_RECORD_TYPE );

    STATE GET_RECORD_TYPE : /* Determine rec. type: garbage, EOList, EOBlock, etc */
      COND( TRUE, gedh_getRecordType, WANT_TO_DELETE);
  
    STATE WANT_TO_DELETE  :  /* Keep processing 'till key match is found */
      COND( IsEOList,    NIL,         UNSUCCESS );     /* We didn't find a match! */
      COND( IsMatched,   NIL,         FOUND_MATCH );   /* This record is a match */
      COND( IsGarbage,   NIL,         GET_NEXT_KV );   /* This is a garbage record */
      COND( IsEOBLK,     gedh_toNxtBlk, LOAD_KV );     /* Record indicates I'm at e.o.block */
      COND( IsUnmatched, NIL,   GET_NEXT_KV );    /* Record was none of above, & it didn't match */
      COND( IsDefault,   NIL,   GET_NEXT_KV );    /* Record was none of above, & it didn't match */
  
    STATE FOUND_MATCH :
      COND( TRUE,   gedh_makeGarb,  QUIT ); /* Turn to garbage and quit. */

    STATE UNSUCCESS :
      return NO_SUCCESS; /* No foundee matchee */
    }

  if (Buffer_change_flag)      /* If current block has changed */
    gedh_write_block();
  return Return_status;
  }

/****************************************************************************
 *
 * NAME: gedh_iterate_disk
 *
 * DESCRIPTION: Successively returns every key/value pair in the hash table on disk.
 * The order of returning pairs is undefined. 
 * 
 * RETURN:  NO_SUCCESS = 0 if there is no values to return, else SUCCESS = 1
 * 
 * CALLED BY: gedhash_iterate
 *
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: Increments htab->disk_iterate and htab->it_pos.
 *               
 * key:   Input: None
 *       output: The next iterative key value found in the hash table
 *
 * keyLen: Input: Length in bytes of key.
 * 
 * val:   Input: None
 *       output: The next iterative val value found in the hash table
 * 
 * valLen: Input: Length in bytes of val.
 *
 ****************************************************************************/
int gedh_iterate_disk(HASH_TAB *htab, char *key, KSIZ *keyLen, char *val, LEN *valLen)
  {
  State state =  LSEEK_TO_IT_POS;  /* First state to start off with */
    

  Return_status = SUCCESS; /* Changed if an I/O error occurs */
  Htab = htab;
  Fh = htab->fh;
  Filepos = 0L;
  Buffer_change_flag = FALSE;
    

  WHILESTATE(state)
    {
    STATE LSEEK_TO_IT_POS :
     
      /* Put pointer past header */
      htab->disk_iterate =
        ((htab->disk_iterate + 1)< (long)HEADER_SIZ) ? (long)HEADER_SIZ : htab->disk_iterate;

      ged_elseek(htab->fh, htab->disk_iterate, SEEK_SET);
      Filepos = htab->disk_iterate;

      /* EOF is not an error here since a read may not be at the start of a block. */
      ged_eread(Fh, Buffer, BLKLEN);
      state =  LOAD_ITERATE_KV;

    STATE LOAD_ITERATE_KV       :  /* Like GET_NEXT_KV, except don't jump to next kv record. */
      COND( IsDefault,                  gedh_loadKv, GET_RECORD_TYPE); 
    
    STATE GET_NEXT_KV   :  /* Jump past 1st cell which contains hash table info */
      COND(IsDefault, gedh_loadNxtKv, GET_RECORD_TYPE);

    STATE LOAD_KV       :  /* Like GET_NEXT_KV, except don't jump to next kv record. */
      COND( IsDefault, gedh_loadKv, GET_RECORD_TYPE );

    STATE GET_RECORD_TYPE :
      COND( IsDefault, gedh_getRecordType, WANT_RECORD);

    STATE WANT_RECORD   :                          /* What kind of record did I just get? */
      COND( IsEOList,    NIL,   WANT_NEXT_ROW );      /* Record indicates I'm at e.o.list */
      COND( IsGarbage,   NIL,   GET_NEXT_KV );    /* This is a garbage record */
      COND( IsEOBLK,     gedh_toNxtBlk, LOAD_KV );     /* Record indicates I'm at e.o.block */
      COND( IsDefault,   NIL,   QUIT );     /* This is a key/val record */

    STATE WANT_NEXT_ROW :
      if (htab->it_pos == htab->h)  /* if no more rows in hash table we are done! */
        {  
        *val = '\0';
        *valLen = 0;
        *key = '\0';
        *keyLen = 0;
        return NO_SUCCESS;  /* No more records to return.  */
        }
      /* Search through all the rows until you find a record */
      htab->disk_iterate = 
        HEADER_SIZ + ( ++(htab->it_pos) * BLKLEN ); /*Point to next row in table */
      state = LSEEK_TO_IT_POS;  
    }

    /* Found a record.  Copy values to function arguments */
    *valLen = Kv->valLen;
    *keyLen = Kv->keyLen;
    memcpy((char *)val, Dval, Kv->valLen);
    memcpy((char *)key, Dkey, Kv->keyLen);

    /* Return_status = SUCCESS; Had a record to return */

    /* Now point to the next record, don't care what kind it is. */
    gedh_loadNxtKv();
    
    /* Reset to point to exact spot of next record */
    htab->disk_iterate = Filepos + ((char *)Kv - (char *)Buffer);
    return Return_status;
    }

/************************************************************************
 * NAME: gedh_dlookup
 *
 * DESCRIPTION: Finds the value of in the disk-based hash table that was
 * last associated with it by gedhash_insert(), and returns the value (and
 * length) via the val and valLen reference parameters.
 *
 * RETURN:  Returns pointer to val, else NULL.  If successful, status is set to
 *         SUCCESS=1, else KEY_NOT_FOUND = -1 or HTAB_NOT_FOUND = -2.

 * CALLED BY: gedhash_lookup
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: None
 *               
 * key:   Input: Look-up key
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 * 
 * val:   Input: None
 *       output: The value associated with key.
 *
 * keyLen: Input: Length in bytes of val.
 *
 * status: Input: None
 *         Output: If successful, SUCCESS = 1, else KEY_NOT_FOUND = -1
 *
 ************************************************************************/
void *  gedh_dlookup(HASH_TAB *htab, void *key, KSIZ keyLen, void *val, LEN *valLen, int *status)
  {
  State state = LSEEK_TO_HASH;  /* First state to start off with */

  /* Initialize global variables */
  Htab = htab;
  Key = (char *)key;
  KeyLen = keyLen;
  Fh = htab->fh;
  Filepos = 0L;
  Buffer_change_flag = FALSE;

  WHILESTATE(state)
    {
    STATE LSEEK_TO_HASH :  /* lseek to right block in hash table & read in value */
      COND(IsDefault, gedh_hash_seek, LOAD_KV);

    STATE GET_NEXT_KV   :  /* Jump to next kv record and read from buffer */
      COND(IsDefault, gedh_loadNxtKv, GET_RECORD_TYPE);

    STATE LOAD_KV       :  /* Like GET_NEXT_KV, except don't jump to next kv record. */
      COND( IsDefault, gedh_loadKv, GET_RECORD_TYPE );

    STATE GET_RECORD_TYPE :
      COND( IsDefault, gedh_getRecordType, WANT_TO_MATCH);

    STATE WANT_TO_MATCH  :                         /* What kind of record did I just get? */
      COND( IsEOList,    NIL,   QUIT        );  /* Record indicates I'm at e.o.list */
      COND( IsMatched,   NIL,   FOUND_MATCH );   /* This record is a match */
      COND( IsGarbage,   NIL,   GET_NEXT_KV ); /* This is a garbage record */
      COND( IsEOBLK,     gedh_toNxtBlk, LOAD_KV );     /* Record indicates I'm at e.o.block */
      COND( IsUnmatched, NIL,   GET_NEXT_KV );   /* Record was none of above, & it didn't match */
      COND( IsDefault,   NIL,   QUIT        );    /* This shouldnever be executed */


    STATE FOUND_MATCH :
      {
      *valLen = Kv->valLen;
      memcpy((char *)val, Dval, Kv->valLen);
      *status = SUCCESS;
      return val; /* success */
      }

    }

  /* Did not find key */
  *(char *)val = (char)NULL;
  *valLen = 0;
  *status = KEY_NOT_FOUND;
  return val;  /* no success. Match not found.  */
  }

/************************************************************************
 * NAME: gedh_insertDisk
 *
 * DESCRIPTION: Counterpart to insertMem, but it inserts a value to the disk
 * The COND() macro executes: "if condition", "perform action", "set next state".
 *
 * CALLED BY:   gedhash_insert
 *
 * RETURN: If successful, returns SUCCESS=1, else zero
 *
 * CALLED BY: gedhash_insert
 * 
 * PARAMETERS:
 * 
 * RETURN VALUE:  Returns zero if unsuccessful, or if a duplicate key is found, 
 * the return value is 2, otherwise 1 is returned.
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: None
 *               
 * key:   Input: Look-up key
 *       output: None
 * 
 * keyLen: Input: Length in bytes of key.
 * 
 * val:   Input: None
 *       output: The value associated with key.
 * 
 * keyLen: Input: Length in bytes of val.
 *
 * status: Input: None
 *         Output: If successful, SUCCESS = 1, else KEY_NOT_FOUND = -1
 *
 ************************************************************************/
int gedh_insertDisk(HASH_TAB *htab, char *key, KSIZ keyLen, char *val, LEN valLen)
  {
  State state = LSEEK_TO_HASH;  /* First state to start off with */

  /* Make sure record length is not greater than block size */
  if (BLKLEN < (sizeof(KVDISK) + keyLen + valLen))
    return NO_SUCCESS;

  /* Initialize global variables */
  Htab = htab;
  Key = key;
  KeyLen = keyLen;
  Val = val;
  ValLen = valLen;
  Fh = htab->fh;
  Garbage.block = 0L;
  Garbage.garb_record = NULL;
  Garbage.size = (LEN)0;
  Filepos = 0L;
  Buffer_change_flag = FALSE;
  IsMatchFound = FALSE;
  Return_status = SUCCESS;

  WHILESTATE(state)
    {
    STATE LSEEK_TO_HASH :  /* lseek to right block in hash table & read in value */
      /* if condition", "perform action", "set next state  */
      COND(IsDefault, gedh_hash_seek, LOAD_KV);

    STATE GET_NEXT_KV   :  /* Jump to next kv record and read from buffer */
      COND(IsDefault, gedh_loadNxtKv, GET_RECORD_TYPE);

    STATE LOAD_KV       :  /* Like GET_NEXT_KV, except don't jump to next kv record. */
      COND( IsDefault, gedh_loadKv, GET_RECORD_TYPE );

    STATE GET_RECORD_TYPE :
      COND( IsDefault, gedh_getRecordType, WANT_TO_INSERT);

    STATE WANT_TO_INSERT  :           /* Keep searching 'till get match or EOList */
      COND( IsEOList,    NIL,           WRITE_EOLIST );  /* Record indicates I'm at e.o.list */
      COND( IsMatched,   SET_DUP_FOUND, FOUND_MATCH );   /* This record is a match */
      COND( IsGarbage,   NIL,           FOUND_GARBAGE ); /* This is a garbage record */
      COND( IsEOBLK,     gedh_toNxtBlk, LOAD_KV );       /* Record indicates I'm at e.o.block */
      COND( IsUnmatched, NIL,   GET_NEXT_KV );   /* Record was none of above, & it didn't match */
      COND( IsDefault,   NIL,   GET_NEXT_KV );    /* Record was none of above, & it didn't match */

    STATE FOUND_GARBAGE :
      /* We had match, but it didn't fit. */
      COND( ((IsMatchFound) && (RightKvSize)), gedh_toBuf,   QUIT );
          
      /* If garbage size is right, set garbage flag */
      COND( RightKvSize,  gedh_setGarb, GET_NEXT_KV );

      COND( IsDefault,    NIL,    GET_NEXT_KV );  /* Garbage was not big enough */

    STATE FOUND_MATCH :
      IsMatchFound = TRUE;                   /* Used in FOUND_GARBAGE state */
      /* The matched record size is same size. Write it here. */
      COND( IsRightSize, gedh_toBuf,     QUIT );

      /* You can fit data here TIGHTER than in the garbage slot */
      COND( IsMatchBest, gedh_toBuf,     QUIT );
      
      /* If data fits into garb, insert, turn Kv into garb. */
      COND( IS_GARBAGE_SLOT,   gedh_makeGarb,  WHERE_IS_GARB ); 
      
      /* No space in block, turn to garb, goto EOLst */
      COND( IsDefault,   gedh_makeGarb,  GET_NEXT_KV ); 

    STATE WHERE_IS_GARB :
      
      /* If garbage is in old block, write it there */
      COND( IsGarbInOldBlock, gedh_toOldBuf,  QUIT );

      /* If garb is in current block, write it there */
      COND( IsGarbInThisBlock, gedh_toGarbBuf,  QUIT );

    STATE WRITE_EOLIST :
      COND( IS_GARBAGE_SLOT,  NIL,     WHERE_IS_GARB ); /* if there's a garbage slot, insert data */
      
      /* if there's room @ end of list, write it and EOL code */
      COND( IsRoomHere, gedh_writeEOL,    QUIT );
      COND( IsDefault,  gedh_writeNewBlk, QUIT ); /* Create a new block & write to it. */
    }
  if (Buffer_change_flag)      /* If current block has changed */
    gedh_write_block();
  return Return_status;
  }

/************************************************************************
 *
 * NAME: gedh_writeNewBlk
 *
 * DESCRIPTION:  Write record to end of list
 *
 * GLOBALS changed: Kv, Filepos, Fh
 *
 ************************************************************************/
void gedh_writeNewBlk(void)
  {
/*  static char newbuf[BLKLEN]; */


  /* make this record an end of blk record */
  Kv->keyLen = (KSIZ)0;
  Kv->valLen = (LEN)EOBLOCK;
  ged_elseek(Fh, Filepos, SEEK_SET);
  Filepos = filelength(Fh);
  Kv->totalLen = filelength(Fh); /* Position of new block */
  ged_ewrite(Fh, Buffer, (unsigned int )BLKLEN);
  Buffer_change_flag = FALSE;

  /* find position of new block at end of file (for debuffing )*/
  ged_elseek(Fh, 0L, SEEK_END);
  Kv = (KVDISK *)Buffer;
  gedh_writeEOL(); /* Write record to end of list */
  }

/************************************************************************
 * 
 * NAME: gedh_writeEOL
 * 
 * DESCRIPTION:  Write record to end of list then create EOL record.
 * 
 * GLOBALS changed: Kv, Buffer_change_flag
 * 
 ************************************************************************/
void gedh_writeEOL(void)
  {
  Kv->totalLen = KeyLen + ValLen;
  gedh_dCopy(Kv, Key, KeyLen, Val, ValLen); /* put data on end of collision list */
  Kv = gedh_kvnext(Kv);
  Kv->keyLen = (KSIZ)0;
  Kv->valLen = (LEN)EOLIST; /* make this record an end of list record */
  Buffer_change_flag = TRUE;
  }


/************************************************************************
 * 
 * NAME: gedh_makeGarb
 * 
 * DESCRIPTION:  Make Kv a garbage record.
 *
 * GLOBALS changed: Kv, Buffer_change_flag
 *
 ************************************************************************/
void gedh_makeGarb(void)
  {
  Kv->keyLen = (KSIZ)0; /* Means this is a special record */
  Kv->valLen = (LEN)GARBAGE; /* Means this is a special GARBAGE record. */
  Buffer_change_flag = TRUE; /* Means I'm going to write to current block before exiting */
  }

/************************************************************************
 * 
 * NAME: gedh_toGarbBuf
 *
 * DESCRIPTION:  Called by gedh_insertDisk(). This is used when we have found
 * a match during an insert operation.  In other words, the key we passed in
 * was already in the hash table.  Normally we write over the matched key/val
 * record, but in this case the record is too small because our new value is
 * larger than the val that is on disk.  If there are any garbage records,
 * the program keeps track of the best fitting garbage slot.  In this case
 * one has been found.  This garbage info is kept in the global,
 * GARB Garbage.garb_record.
 * 
 * What this function does 1st is to turn the matched record into a garbage
 * record (why? because our key/val would not fit into it).  It then
 * copies our key/val into garbage record, which is in current block.
 *
 * GLOBALS changed:  Garbage.garb_record
 * 
 ************************************************************************/
void gedh_toGarbBuf(void)
  {
  if (Kv->keyLen) /* If found match which does not fit, turn into garbage */
    gedh_makeGarb(); /* Make new garbage record */
  gedh_dCopy(Garbage.garb_record, Key, KeyLen, Val, ValLen);
  Buffer_change_flag = TRUE;
  }

/************************************************************************
 *
 * NAME: gedh_toOldBuf
 *
 * DESCRIPTION:  Called by gedh_insertDisk(). This is used when we have found
 * a match during an insert operation.  In other words, the key we passed in
 * was already in the hash table.  Normally we write over the matched key/val
 * record, but in this case the record is too small because our new value is
 * larger than the val that is on disk.  If there are any garbage records,
 * the program keeps track of the best fitting garbage slot.  In this case
 * one has been found.  This garbage info is kept in the global,
 * GARB Garbage.garb_record.
 *
 * What this function does 1st is to turn the matched record into a garbage
 * record (why? because our key/val would not fit into it).  It then
 * copies our key/val into garbage record, which is in another block.
 *
 * NOTE:  This needs to be tested!  Make sure current Kv is being turned into a
 * garbage record.
 *
 * GLOBALS changed:  Garbage.garb_record
 *
 ************************************************************************/
void gedh_toOldBuf(void)
  {
  char garBuf[BLKLEN];

  /* Now seek to garbage record's block & insert data there */
  ged_elseek(Fh, Garbage.block, SEEK_SET); /* Put position to beginning of garbage block */
  gedh_read_block(garBuf); /* read a block from disk into Buffer */
  Garbage.garb_record = (KVDISK *)((char *)(Garbage.garb_record) - Buffer + garBuf);
  gedh_dCopy(Garbage.garb_record, Key, KeyLen, Val, ValLen); /* Copy data into garbage record */
  ged_elseek(Fh, Garbage.block, SEEK_SET); /* Put position to beginning of garbage block */
  ged_ewrite(Fh, garBuf, (unsigned int)BLKLEN); /* write to block */
  ged_elseek(Fh, Filepos, SEEK_SET); /* Return to original position  */
  if (Kv->keyLen) /* If found match which does not fit, turn into garbage */
     gedh_makeGarb(); /* We don't turn a special record (i.e. keyLen == 0) to garb */
  }

/************************************************************************
 *
 * NAME: gedh_setGarb
 *
 * DESCRIPTION:  Sets the garbage struct to the best garbage slot found so far.
 * It is assumed that we know that Garbage.Size & Kv is greater than
 * (KeyLen + ValLen).
 *
 ************************************************************************/
void gedh_setGarb(void)
  {
  /* If Garbage has not been set yet, Garbage will now be Kv.  If Garbage has
   * been set then Garbage will now be the tightest fit between Kv & Garbage. */

  if (!IS_GARBAGE_SLOT || (IS_GARBAGE_SLOT && gedh_closer_fit()))
    {
    Garbage.block = Filepos; /* block this garbage is in. */
    Garbage.garb_record = Kv; /* Points to the position within the block */
    Garbage.size = gedh_kvsize(Kv); /* Size of garbage record */
    }
  }

/************************************************************************
 * 
 * NAME: gedh_closer_fit
 * 
 * DESCRIPTION:  Determines if the current garbage candidate has a closer
 * fit than the garbage record found previously.  As I traverse a collision
 * list looking for a match or a place to insert a key & value, I keep track
 * of potential garbage slots in case I can insert them there.  It is assumed
 * that we know that Garbage.Size & Kv is greater than (KeyLen + ValLen). Kv is
 * the current garbage record that we are considering changing to
 *
 ************************************************************************/
GHBool gedh_closer_fit(void)
  {
  return ((gedh_kvsize(Kv) - (KeyLen + ValLen)) < (Garbage.size - (KeyLen + ValLen)));
  }


/************************************************************************
 * 
 * NAME: gedh_toBuf
 * 
 * DESCRIPTION:  Copies to current position in buffer. Buffer_change_flag is set
 *
 ************************************************************************/
void gedh_toBuf(void)
  {
  gedh_dCopy(Kv, Key, KeyLen, Val, ValLen); /* put val on new block */
  Buffer_change_flag = TRUE;
  }  


/************************************************************************
 * 
 * NAME: gedh_toNxtBlk
 *
 * DESCRIPTION:  I have hit an end of block.  This will put me onto the next
 * one.  It will also write the buffer to the current block if there were
 * any changes.
 *
 ************************************************************************/
void gedh_toNxtBlk(void)
  {
  if (Buffer_change_flag)      /* If current block has changed */
    {
    ged_elseek(Fh, Filepos, SEEK_SET);
     ged_ewrite(Fh, Buffer, (unsigned int)BLKLEN); /* write to block. */
    Buffer_change_flag = FALSE;
    }
  ged_elseek(Fh, Kv->totalLen, SEEK_SET); /* jump to next block */
  Filepos = Kv->totalLen;      /* save the current file position */
  gedh_read_block(Buffer);
  }

/************************************************************************
 * 
 * NAME: gedh_loadKv  DESCRIPTION: loads Buffer into Kv.
 *
 ************************************************************************/
void gedh_loadKv(void)
  {
  Kv = (KVDISK *)Buffer;
  gedh_kvGet(Kv, &Dkey, &Dval); /* load in values from disk buffer */
  }

/************************************************************************
 *
 * NAME: gedh_loadNxtKv  DESCRIPTION: Increments Kv past the current KVPAIR
 * struct record in Buffer and loads the next record into Kv.
 *
 ************************************************************************/
void gedh_loadNxtKv(void)
  {
  Kv = gedh_kvnext(Kv);
  gedh_kvGet(Kv, &Dkey, &Dval); /* load in values from disk buffer */
  }

/************************************************************************
 * 
 * NAME: gedh_hash_seek
 *
 * DESCRIPTION: Determines hash value and positions file pointer to the
 *              correct block within the hash table file.
 *
 ************************************************************************/
void gedh_hash_seek(void)
  {
  ged_elseek(Fh, (Filepos = HEADER_SIZ + 
     (gedh_hash(Key, KeyLen, Htab->h) * BLKLEN)), SEEK_SET);
  gedh_read_block(Buffer);
  }

/************************************************************************
 * 
 * NAME: gedh_write_block
 *
 * DESCRIPTION: Reads a block from disk into Buffer
 *
 ************************************************************************/
void gedh_write_block(void)
  {
  ged_elseek(Fh, Filepos, SEEK_SET); /* Put position to beginning of block */
  ged_ewrite(Fh, Buffer, (unsigned int)BLKLEN); /* write to block. */
  Buffer_change_flag = FALSE;
  }

/************************************************************************
 * 
 * NAME: gedh_read_block
 *
 * DESCRIPTION: Reads a block from disk into Buffer.  The buf argument passed
 * in points to the global, Buffer.
 *
 * CALLED BY: gedh_toNxtBlk, gedh_hash_seek
 *  
 ************************************************************************/
void gedh_read_block(char *buf) { ged_eread(Fh, buf, BLKLEN); }

/************************************************************************
 * 
 * NAME: gedh_kvsize
 *
 * DESCRIPTION: Returns the size of the KVDISK record. Size only includes
 * the size of val and key.  
 *              
 * RETURN: kv->totalLen, which is total size of the KVDISK record.
 * 
 ************************************************************************/
LEN gedh_kvsize(KVDISK *kv) { return (LEN)(kv->totalLen); }

/************************************************************************
 * NAME: gedh_getRecordType
 *
 * DESCRIPTION: Sets RecordType to the records attribute: EOLIST, GARBAGE,
 * EOBLOCK, MATCHED, or UNMATCHED
 *
 * CALLED BY: gedh_delete_disk, gedh_iterate_disk, gedh_dlookup, 
 *            gedh_insertDisk
 ************************************************************************/
void gedh_getRecordType(void)
  {
  if (Kv->keyLen == 0) /* If this record is EOLine or EOBlock or garbage */
    {
    if ( (Kv->valLen > (LEN)2) || (Kv->valLen < (LEN)0) ) 
      ged_error( RECORD_ERR, NULL ); /* Fatal error! Unidentified record type. */
    else
      RecordType = Kv->valLen; /* Return record type */
    }
  else if ((KeyLen == Kv->keyLen) && (memcmp(Key, Dkey, (size_t)KeyLen))==0)
    RecordType = MATCHED; /* found match for Key */
  else RecordType = UNMATCHED; /* Key match not found */
  }

    
/****************************************************************************
 *
 * NAME: ged_eclose
 *
 * DESCRIPTION: Executes close() and performs error-checking
 * 
 * RETURN:  0 if successful, -1 if not. The global, Return_status, is set to
 *          NO_SUCCESS
 * 
 * PARAMETERS: Same as the standard close() function
 ****************************************************************************/
int ged_eclose(int handle)
  {
  int success = close(handle);

  if (success == -1)
    {  
    ged_error(CLOSE_ERR, NULL);
    Return_status = NO_SUCCESS;
    }
  return success;
  }    
    

/****************************************************************************
 *
 * NAME: ged_ewrite
 *
 * DESCRIPTION: Same as an write with a call to ged_error if failure. The "e"
 * in "e"write stands for error.
 * 
 * RETURN:  Returns the number of characters written, or -1 (IO_ERROR)
 *          if no success. If no success, the global Return_status, is
 *          set to NO_SUCCESS.
 *
 * CALLS: ged_error() if not successful.
 *
 * PARAMETERS: Same as the standard write() function
 * 
 ****************************************************************************/
int ged_ewrite(int handle, void *local_buf, unsigned int count)
  {
  int result =  write(handle, local_buf, count);
  if (result == -1)
    {  
    ged_error(WRITE_ERR, NULL); /* Could not write to file. */
    Return_status = NO_SUCCESS;
    }
  return result;
  }

/****************************************************************************
 *
 * NAME: ged_eopen
 *
 * DESCRIPTION: Same as an open with a call to ged_error if failure. The "e"
 *              in "e"open stands for error. Returns pointer to the opened file.
 * 
 * RETURN:  A handle if successful. If not -1 is returned and the global,
 *          Return_status, is set to NO_SUCCESS.
 *
 * PARAMETERS:  Same as the standard open() function.  The pmode argument is
 * not optional.
 * 
 ****************************************************************************/
int ged_eopen(char *filename, int oflag, int pmode)
  {
  int fh; /* File handle */

  if (( fh = open(filename, oflag, pmode)) == -1)
    {  
    ged_error(OPEN_ERR, NULL); /* Could not open file */
    Return_status = NO_SUCCESS;
    }
  return fh;
  }

/****************************************************************************
 *
 * NAME: ged_efopen
 *
 * DESCRIPTION: Same as an fopen with a call to ged_error if failure. The "e"
 *              in "e"fopen stands for error. Returns pointer to the opened file.
 * 
 * RETURN:  A pointer to the open file if successful, else NULL and the global,
 *          Return_status, is set to NO_SUCCESS.
 *
 * PARAMETERS:  Same as the standard fopen() function
 * 
 *   path:      input: Full path of file to open
 *              output: none
 *
 *   mode:      input: Type of access permitted.
 *              ouput: None.
 * 
 *****************************************************************************/
FILE *ged_efopen(char *path, char *mode)
  {
  FILE *fp;

  if (!(fp = fopen(path, mode)))
    {  
    ged_error(FOPEN_ERR, NULL); /* Could not open file */
    Return_status = NO_SUCCESS;
    }
  return fp;
  }

/****************************************************************************
 *
 * NAME: ged_eread
 *
 * DESCRIPTION: Executes read() and performs error checking
 * 
 * RETURNS:  The number of bytes actually read or 0 at end-of-file, if
 *           successful, or -1 if not.  (errno: EBADF)
 * 
 * PARAMETERS: Same as the standard read() function
 * 
 ****************************************************************************/
int ged_eread( int handle, void *buffer, unsigned count )
  {
  int success = read(handle, buffer, count);

  if (success == -1)
    {  
    ged_error(READ_ERR, NULL);
    Return_status = NO_SUCCESS;
    }
  return success;
  }


/****************************************************************************
 *
 * NAME: ged_elseek
 *
 * DESCRIPTION: Performs lseek operation and checks for error.
 * 
 * RETURN: IO_SUCCESS = 1 or IO_ERROR = -1
 * 
 * PARAMETERS: Same as the standard lseek() function
 *  
 ****************************************************************************/
int ged_elseek( int handle, long offset, int origin)
  {  
  if (lseek(handle, offset, origin) == -1L)
    {  
    ged_error(SEEK_ERR, (NODE *)NULL); /* Seek error */
    Return_status = NO_SUCCESS;
    return IO_ERROR;
    }
  return IO_SUCCESS;
  }
    
/****************************************************************************
 *
 * NAME: gedh_save_htab_struct
 *
 * DESCRIPTION: Saves a hash table's header in the first cell in the hash table.
 *
 * RETURN:  SUCCESS (1) upon success, NO_SUCCESS (zero) otherwise.
 * 
 * CALLED BY: gedhash_close()
 * 
 * PARAMETERS:
 * 
 * htab:  input: Pointer to HASH_TAB struct.
 *       output: None
 *
 ****************************************************************************/
int gedh_save_htab_struct(HASH_TAB *htab)
  {
  if ( !htab || !htab->fh )
    return NO_SUCCESS;
  if ( (ged_elseek(htab->fh, 15, SEEK_SET) == IO_ERROR)
     || (ged_ewrite(htab->fh, htab, sizeof(HASH_TAB)) == IO_ERROR) )
    return NO_SUCCESS;
  return SUCCESS;
  }
