/****************************************************************************
 *      (c) 1991    Corporation of the President of the CHURCH of JESUS CHRIST
 *      Copyright   of LATTER-DAY SAINTS.
 *      All rights reserved.  Unauthorized reproduction of this software is
 *      prohibited and is in violaton of United States copyright laws.
 ****************************************************************************/

/****************************************************************************
 *   FILE:   GEDPOINT.C
 *
 *   DESCRIPTION: Contains one function, gedhash_create_xref() which creates
 *   a hash table and stores all XREF's found in ged_filename.
 *
 *   AUTHOR: Larry A. Maddocks
 *
 *   DATE:   May 17, 1991
 * 
 *   MODIFICATION HISTORY:
 * ----------------------------------------------------------------------
 * |  Date            | Who | Comments
 * ----------------------------------------------------------------------
 * | 08/07/91 08:53am | LAM | Added code to handle multi-disk gedcom files.
 * ----------------------------------------------------------------------
 *
 ****************************************************************************/

#include "gedcom.h"
#include "GEDHSTAT.H"

/****************************************************************************
 *
 * NAME: gedhash_create_xref
 *
 * DESCRIPTION:  Creates a hash table then makes one pass over a gedcom file,
 * storing the lseek position of each record which contains an xref id. Spans
 * Multiple disks until a TRLR record is found. A state machine is used in
 * this function.
 *
 * NOTE:  This must be revised to handle GEDOCOM that goes more than 9 levels deep.
 *
 * CALLED BY:  gedhash_create_xref calls fall mainly in your main().
 *
 * PARAMETERS:
 *
 * ged_filename:  input: Name of GEDCOM file containing the xref's.
 *
 * hash_filename: input: Name of file that will store the hash table.  If the
 *                hash table is to be memory-based, this argument must be NULL.
 * 
 * h_size:        input: Size of hash table.
 *                        
 ****************************************************************************/
HASH_TAB gedhash_create_xref(char *ged_filename, char *hash_filename, HSIZ h_size)
  {
  CX_State state = WANT_LEVEL; /* Inititalize First state */
  HASH_TAB htab; /* Struct created & returned by this function. */
  int      c;     /* Gets each character that's read from ged_filename */
  char     key[MAX_XREF_ID_LENGTH + 1]; /* Key stored in the hash table */
  int      keyLen = 0; /* Length of key.  Stored in hash table */
  int      trlr = 0;   /* Counter used in WANT_TERMINAT_OR_TRLR_TAG state. */
  char     trlr_tag[5]; /* Used in WANT_TERMINAT_OR_TRLR_TAG state. */
  SEEK     seek;       /* Value stored in hash table.  See comment in SEEK struct declaration */

  strcpy(trlr_tag, "TRLR");
  if (gedhash_create_table(&htab, hash_filename, h_size, MAX_MEMORY)
      ==(HASH_TAB *)NULL) /* Create hash tabel */
    ged_error(CREATE_TABLE_ERR, (NODE *)NULL); /* Could not create hash table.  */

  /* Initialize gedcom file extention thingers for multiple-file gedcom */
  seek.ged_ext = -1; /* Means this file's extention is .GED */
  gedhash_fopen(&htab, ged_filename, "rb"); /* Open htab.fg and set htab.ged_ext to -1 */

  /* Now make a pass and insert all pointers */
  for (;;) /* for ever */
    {
    while ((c = getc(htab.fg)) != EOF)
      {
      switch (state)
        {
        when  WANT_LEVEL :
          if (isdigit(c))
            {
            seek.seek_position = ftell(htab.fg) - 1L;
            state = WANT_1ST_AT_SIGN_OR_TAG;
            }
        when  WANT_1ST_AT_SIGN_OR_TAG :
          if (c == '@')
            state = WANT_XREF;
          else if (isalpha(c))
            {
            if (c == trlr_tag[trlr])
              {
              trlr++;
              state = WANT_TERMINAT_OR_TRLR_TAG;
              }
            else
              state = WANT_TERMINATOR;
            }

        when  WANT_XREF :
          if (c != '@') /* last @ sign */
            {
            key[keyLen] = (char)c; /* Copy xref character into key[] */
            keyLen++;
            if (keyLen > MAX_XREF_ID_LENGTH) /* keyLen is greater than MAX_XREF_ID_LENGTH.  */
              ged_error( XREF_LENGTH_ERR, (NODE *)NULL);
            }
          else
            {
            if (gedhash_insert(&htab, key, (KSIZ)keyLen, &seek, (LEN)sizeof(SEEK)) != SUCCESS)
              ged_error( INS_ERR, (NODE *)NULL); /* Could not insert value into hash table */
            keyLen = 0; /* reset */
            state = HAVE_KEY_WANT_TERMINATOR;
            }

        when  WANT_TERMINATOR :
          if (c == '\n' || c == '\r')
            state = WANT_LEVEL;

        when WANT_TERMINAT_OR_TRLR_TAG :
          if (c == '\n' || c == '\r' )
            {
            state = WANT_LEVEL;
            trlr = 0;  /* Reset counter */
            }
          else if (c != trlr_tag[trlr])
            {
            state = WANT_TERMINATOR;
            trlr = 0;  /* Reset counter */
            }
          else if (trlr++ == 3)
            return htab;             /* The only return is here */

        when HAVE_KEY_WANT_TERMINATOR :
          if (c == '\n'  || c== '\r' )
            state = WANT_LEVEL;
        }
      }
    gedh_floppy_prompt(&htab, ++(seek.ged_ext)); /* Check for change in multidisk gedcom file */
    }
    ged_error( NO_TRLR_ERR, (NODE *)NULL); /* No TRLR tag encountered.  */
  }
